#include "query.h"
#include "scope.h"
#include "preview.h"
#include <json/reader.h>
#include <json/writer.h>

#include <fstream>
#include <iostream>

namespace sc = unity::scopes;
using namespace std;
using namespace scope;

void Scope::start(string const& scope_id) {
    setlocale(LC_ALL, "");
    scope_dir = scope_directory();
    load_config(scope_id);
}

void Scope::stop() {
}

sc::Variant to_variant(Json::Value const& value)
{
    switch (value.type())
    {
        case Json::ValueType::nullValue:
            return sc::Variant::null();
        case Json::ValueType::arrayValue:
            {
                sc::VariantArray arr;
                for (unsigned int i=0; i<value.size(); ++i)
                {
                    arr.push_back(to_variant(value[i]));
                }
                return sc::Variant(arr);
            }
        case Json::ValueType::objectValue:
            {
                sc::VariantMap var;
                for (auto const& m: value.getMemberNames())
                {
                    var[m] = to_variant(value[m]);
                }
                return sc::Variant(var);
            }
        case Json::ValueType::stringValue:
            return sc::Variant(value.asString());
        case Json::ValueType::intValue:
        case Json::ValueType::uintValue:
            // this can throw std::runtime_error from jsoncpp if unsigned int to int conversion is not possible
            if (value.isInt())
            {
                return sc::Variant(value.asInt());
            }
            return sc::Variant(static_cast<int64_t>(value.asInt64()));
        case Json::ValueType::realValue:
            return sc::Variant(value.asDouble());
        case Json::ValueType::booleanValue:
            return sc::Variant(value.asBool());
        default:
            {
                throw std::runtime_error("Unsupported json type " + std::to_string(static_cast<int>(value.type())));
            }
    }
}

void Scope::load_config(const std::string& scope_id)
{
    Json::Value root;
    const std::string cfgfile = scope_directory() + "/" + scope_id + ".json";
    Json::Reader reader;
    Json::FastWriter writer; // used to create renderer string in json format
    std::ifstream file(cfgfile);
    reader.parse(file, root);
    if (root.isObject()) {
        if (root.isMember("categories")) {
            auto const catsArray = root["categories"];
            if (catsArray.isArray()) {
                for (unsigned i = 0; i<catsArray.size(); i++) {
                    auto const& catsDict = catsArray[i];
                    if (!catsDict.isObject()) {
                        throw std::runtime_error("'categories' element is not an object, scope " + scope_id);
                    }

                    CategoryDefinition cat_def;

                    if (catsDict.isMember("id")) {
                        cat_def.id = catsDict["id"].asString();
                    } else {
                        throw std::runtime_error("Missing category id, scope " + scope_id);
                    }
                    if (catsDict.isMember("name")) {
                        cat_def.name = catsDict["name"].asString();
                    } else {
                        throw std::runtime_error("Missing category name, scope " + scope_id);
                    }
                    if (catsDict.isMember("renderer")) {
                        auto rendererObj = catsDict["renderer"];
                        if (!rendererObj.isObject()) {
                            throw std::runtime_error("'renderer' element is not an object, scope " + scope_id);
                        }
                        cat_def.renderer = writer.write(rendererObj);
                    } else {
                        throw std::runtime_error("Missing renderer definition, scope " + scope_id);
                    }
                    if (catsDict.isMember("results")) {
                        auto resultsObj = catsDict["results"];
                        if (!resultsObj.isObject()) {
                            throw std::runtime_error("'results' element is not an object, scope " + scope_id);
                        }
                        if (resultsObj.isMember("count")) {
                            cat_def.result_count = resultsObj["count"].asInt();
                        } else {
                            throw std::runtime_error("Missing result count, scope " + scope_id + ", category " + cat_def.id);
                        }
                        if (!resultsObj.isMember("result_template")) {
                            throw std::runtime_error("Missing icon uri pattern, scope " + scope_id + ", category " + cat_def.id);
                        }

                        // Handle result_template
                        auto resultTemplObj = resultsObj["result_template"];
                        if (!resultTemplObj.isObject()) {
                            throw std::runtime_error("'result_template' element is not an object, scope " + scope_id);
                        }
                        for (auto const& mem: resultTemplObj.getMemberNames()) {
                            cat_def.result_values.push_back(std::make_pair<>(mem, to_variant(resultTemplObj[mem])));
                        }
                    } // it's ok to have no results
                    category_defs.push_back(cat_def);

                } // end of category array processing loop
            } else {
                throw std::runtime_error("'categories' element is not an array, scope " + scope_id);
            }
        } else {
            throw std::runtime_error("No categories present for scope " + scope_id);
        }
    } else {
        throw std::runtime_error("Root element of the config file " + cfgfile + " is not an object");
    }
}

sc::SearchQueryBase::UPtr Scope::search(const sc::CannedQuery &query,
                                        const sc::SearchMetadata &metadata) {
    // Boilerplate construction of Query
    return sc::SearchQueryBase::UPtr(new Query(*this, query, metadata));
}

sc::PreviewQueryBase::UPtr Scope::preview(sc::Result const& result,
                                          sc::ActionMetadata const& metadata) {
    // Boilerplate construction of Preview
    return sc::PreviewQueryBase::UPtr(new Preview(result, metadata));
}

#define EXPORT __attribute__ ((visibility ("default")))

// These functions define the entry points for the scope plugin
extern "C" {

EXPORT
unity::scopes::ScopeBase*
// cppcheck-suppress unusedFunction
UNITY_SCOPE_CREATE_FUNCTION() {
    return new Scope();
}

EXPORT
void
// cppcheck-suppress unusedFunction
UNITY_SCOPE_DESTROY_FUNCTION(unity::scopes::ScopeBase* scope_base) {
    delete scope_base;
}

}
